"""ivf metadata ordering is a module to reorder a metadata collection by ivf clusters"""

import os
from tqdm import tqdm
from pathlib import Path
import numpy as np
from collections import defaultdict
import heapq
import time
import pandas as pd

import pyarrow.parquet as pq
import h5py
import faiss


def search_to_new_ids(index, query, k):
    """
    this function maps the result ids to the ones ordered by the ivf clusters
    to be used along with a re-ordered metadata
    """
    distances, indices = index.search(query, k)
    opq2 = faiss.downcast_VectorTransform(index.chain.at(0))
    xq = opq2.apply(query)
    _, l = faiss.extract_index_ivf(index).quantizer.search(xq, faiss.extract_index_ivf(index).nprobe)
    il = faiss.extract_index_ivf(index).invlists
    list_sizes = [il.list_size(i) for i in range(il.nlist)]
    starting_offset = []
    c = 0
    for i in list_sizes:
        starting_offset.append(c)
        c += i
    old_id_to_new_id = {}
    for i in l[0]:
        i = int(i)
        ids = il.get_ids(i)
        list_size = il.list_size(int(i))
        items = faiss.rev_swig_ptr(ids, list_size)
        for nit, it in enumerate(items):
            old_id_to_new_id[it] = starting_offset[i] + nit
        il.release_ids(ids=ids, list_no=i)
    ids = np.array([old_id_to_new_id[i] if i != -1 else -1 for i in indices[0]])
    return distances, ids


def get_old_to_new_mapping(index):
    """
    use an ivf index to compute a mapping from initial ids to ids ordered by clusters
    """
    il = faiss.extract_index_ivf(index).invlists
    d = np.ones((index.ntotal,), "int64")
    begin_list = []
    current_begin = 0
    for i in tqdm(range(il.nlist)):
        begin_list.append(current_begin)
        ids = il.get_ids(i)
        list_size = il.list_size(int(i))
        items = faiss.rev_swig_ptr(ids, list_size)
        new_ids = range(current_begin, current_begin + list_size)
        d.put(np.array(items, "int"), np.array(new_ids, "int"))
        il.release_ids(ids=ids, list_no=i)
        current_begin += list_size

    return d


def re_order_parquet(index, input_path, output_path, columns_to_return):
    """
    use external sort to reorder parquet files
    """
    d = get_old_to_new_mapping(index)
    data_dir = Path(input_path)
    if not os.path.exists(output_path):
        os.mkdir(output_path)
    current_offset = 0
    current_id = 0
    for parquet_files in tqdm(sorted(data_dir.glob("*.parquet"))):
        df = pd.read_parquet(parquet_files)
        df["new_id"] = d[current_offset : current_offset + len(df)]
        saved_df = df[columns_to_return + ["new_id"]]
        saved_df = saved_df.sort_values("new_id")
        saved_df.to_parquet(output_path + "/meta_" + str(current_id) + ".parquet")
        current_id += 1
        current_offset += len(df)


class Hdf5Sink:
    """
    A hdf5 sink: take as input rows and write them to hdf5 regularly
    """

    def __init__(self, output_hdf5_file, keys):
        self.f = h5py.File(output_hdf5_file, "w")
        self.ds = self.f.create_group("dataset")
        self.buffer = []
        self.keys = keys

    def write(self, sample):
        self.buffer.append(sample)
        if len(self.buffer) == 10**6:
            self._write_buffer()

    def end(self):
        self._write_buffer()
        self.f.close()

    def _write_buffer(self):
        """
        Write a list of rows to hdf5
        """
        if len(self.buffer) == 0:
            return
        df = pd.DataFrame(self.buffer, columns=self.keys)
        for k, v in df.items():
            if k not in self.keys:
                continue
            col = v
            if col.dtype in ("float64", "float32"):
                col = col.fillna(0.0)
            if col.dtype in ("int64", "int32"):
                col = col.fillna(0)
            if col.dtype == "object":
                col = col.fillna("")
            z = col.to_numpy()
            if k not in self.ds:
                self.ds.create_dataset(k, data=z, maxshape=(None,), compression="gzip")
            else:
                prevlen = len(self.ds[k])
                self.ds[k].resize((prevlen + len(z),))
                self.ds[k][prevlen:] = z
        self.buffer = []


class DummySink:
    def __init__(self):
        pass

    def write(self, sample):
        pass

    def end(self):
        pass


def external_sort_parquet(output_sink, input_path):
    """
    create heap
    add to heap 1 batch of each file
    store in dict nb of item in heap for each file
    start getting from heap and pushing to sink
    when nb_item[last_retrieved] == 0 and there is some item left in this file, add a new batch of that file in heap
    """

    h = []
    data_dir = Path(input_path)
    files = [pq.ParquetFile(filename, memory_map=True) for filename in sorted(data_dir.glob("*.parquet"))]
    batches_list = [ffile.iter_batches(batch_size=10**4) for ffile in files]
    index_to_value = {}
    counts = [ffile.metadata.num_rows for ffile in files]
    current_count_per_file = defaultdict(lambda: 0)

    def read_batch(i):
        batch = next(batches_list[i])
        current_count_per_file[i] += batch.num_rows
        df = batch.to_pandas()
        data = zip(df["new_id"], *[df[c] for c in [c for c in df.columns if c != "new_id"]])
        for e in data:
            heapq.heappush(h, (e[0], i))
            index_to_value[e[0]] = e[1:]

    for i in range(len(batches_list)):
        read_batch(i)

    done_count_per_file = defaultdict(lambda: 0)
    c = 0
    begin = time.time()
    while h:
        c += 1
        e, i = heapq.heappop(h)
        v = index_to_value[e]
        del index_to_value[e]
        output_sink.write(v)
        current_count_per_file[i] -= 1
        done_count_per_file[i] += 1
        if current_count_per_file[i] == 0 and done_count_per_file[i] < counts[i]:
            read_batch(i)
        if c % 100000 == 0:
            print(e, c, time.time() - begin, "s")

    output_sink.end()



"""distributors provide way to compute using several gpus and several machines"""

import os

from .worker import worker


class SequentialDistributor:
    def __init__(self, tasks, worker_args):
        self.tasks = tasks
        self.worker_args = worker_args

    def __call__(self):
        """
        call a single `worker(...)` and pass it everything.
        """
        worker(
            tasks=self.tasks,
            **self.worker_args,
        )


class PysparkDistributor:
    """the pyspark distributor uses pyspark for distribution"""

    def __init__(self, tasks, worker_args):
        self.tasks = tasks
        self.worker_args = worker_args

    def __call__(self):
        """
        Parallelize work and call `worker(...)`
        """

        import pyspark  # pylint: disable=import-outside-toplevel
        from pyspark.sql import SparkSession  # pylint: disable=import-outside-toplevel

        spark = SparkSession.getActiveSession()

        if spark is None:
            print("No pyspark session found, creating a new one!")
            spark = (
                SparkSession.builder.config("spark.driver.memory", "16G")
                .master("local[" + str(2) + "]")
                .appName("spark-stats")
                .getOrCreate()
            )

        rdd = spark.sparkContext.parallelize(c=self.tasks, numSlices=len(self.tasks))

        def run(partition_id):
            context = pyspark.TaskContext.get()
            if "gpu" in context.resources():
                gpu = context.resources()["gpu"].addresses[0]
                os.environ["CUDA_VISIBLE_DEVICES"] = gpu

            worker(tasks=[partition_id], **self.worker_args)

        rdd.foreach(run)



"""load clip"""
from functools import lru_cache
from torch import autocast, nn
import torch
import clip
from PIL import Image
import time
import numpy as np


class HFClipWrapper(nn.Module):
    """
    Wrap Huggingface ClipModel
    """

    def __init__(self, inner_model, device):
        super().__init__()
        self.inner_model = inner_model
        self.device = torch.device(device=device)
        if self.device.type == "cpu":
            self.dtype = torch.float32
        else:
            self.dtype = torch.bfloat16 if torch.cuda.is_bf16_supported() else torch.float16

    def encode_image(self, image):
        if self.device.type == "cpu":
            return self.inner_model.get_image_features(image.squeeze(1))
        with autocast(device_type=self.device.type, dtype=self.dtype):
            return self.inner_model.get_image_features(image.squeeze(1))

    def encode_text(self, text):
        if self.device.type == "cpu":
            return self.inner_model.get_text_features(text)
        with autocast(device_type=self.device.type, dtype=self.dtype):
            return self.inner_model.get_text_features(text)

    def forward(self, *args, **kwargs):
        return self.inner_model(*args, **kwargs)


class OpenClipWrapper(nn.Module):
    """
    Wrap OpenClip for managing input types
    """

    def __init__(self, inner_model, device):
        super().__init__()
        self.inner_model = inner_model
        self.device = torch.device(device=device)
        if self.device.type == "cpu":
            self.dtype = torch.float32
        else:
            self.dtype = torch.bfloat16 if torch.cuda.is_bf16_supported() else torch.float16

    def encode_image(self, image):
        if self.device.type == "cpu":
            return self.inner_model.encode_image(image)
        with autocast(device_type=self.device.type, dtype=self.dtype):
            return self.inner_model.encode_image(image)

    def encode_text(self, text):
        if self.device.type == "cpu":
            return self.inner_model.encode_text(text)
        with autocast(device_type=self.device.type, dtype=self.dtype):
            return self.inner_model.encode_text(text)

    def forward(self, *args, **kwargs):
        return self.inner_model(*args, **kwargs)


def load_hf_clip(clip_model, device="cuda"):
    """load hf clip"""
    from transformers import CLIPProcessor, CLIPModel  # pylint: disable=import-outside-toplevel

    model = CLIPModel.from_pretrained(clip_model)
    preprocess = CLIPProcessor.from_pretrained(clip_model).image_processor
    model = HFClipWrapper(inner_model=model, device=device)
    model.to(device=device)
    return model, lambda x: preprocess(x, return_tensors="pt").pixel_values


def load_open_clip(clip_model, use_jit=True, device="cuda", clip_cache_path=None):
    """load open clip"""

    import open_clip  # pylint: disable=import-outside-toplevel

    torch.backends.cuda.matmul.allow_tf32 = True

    pretrained = dict(open_clip.list_pretrained())
    checkpoint = pretrained[clip_model]
    model, _, preprocess = open_clip.create_model_and_transforms(
        clip_model, pretrained=checkpoint, device=device, jit=use_jit, cache_dir=clip_cache_path
    )
    model = OpenClipWrapper(inner_model=model, device=device)
    model.to(device=device)
    return model, preprocess


class DeepSparseWrapper(nn.Module):
    """
    Wrap DeepSparse for managing input types
    """

    def __init__(self, model_path):
        super().__init__()

        import deepsparse  # pylint: disable=import-outside-toplevel

        ##### Fix for two-input models
        from deepsparse.clip import CLIPTextPipeline  # pylint: disable=import-outside-toplevel

        def custom_process_inputs(self, inputs):
            if not isinstance(inputs.text, list):
                # Always wrap in a list
                inputs.text = [inputs.text]
            if not isinstance(inputs.text[0], str):
                # If not a string, assume it's already been tokenized
                tokens = np.stack(inputs.text, axis=0, dtype=np.int32)
                return [tokens, np.array(tokens.shape[0] * [tokens.shape[1] - 1])]
            else:
                tokens = [np.array(t).astype(np.int32) for t in self.tokenizer(inputs.text)]
                tokens = np.stack(tokens, axis=0)
                return [tokens, np.array(tokens.shape[0] * [tokens.shape[1] - 1])]

        # This overrides the process_inputs function globally for all CLIPTextPipeline classes
        CLIPTextPipeline.process_inputs = custom_process_inputs
        ####

        self.textual_model_path = model_path + "/textual.onnx"
        self.visual_model_path = model_path + "/visual.onnx"

        self.textual_model = deepsparse.Pipeline.create(task="clip_text", model_path=self.textual_model_path)
        self.visual_model = deepsparse.Pipeline.create(task="clip_visual", model_path=self.visual_model_path)

    def encode_image(self, image):
        image = [np.array(image)]
        embeddings = self.visual_model(images=image).image_embeddings[0]
        return torch.from_numpy(embeddings)

    def encode_text(self, text):
        text = [t.numpy() for t in text]
        embeddings = self.textual_model(text=text).text_embeddings[0]
        return torch.from_numpy(embeddings)

    def forward(self, *args, **kwargs):  # pylint: disable=unused-argument
        return NotImplemented


def load_deepsparse(clip_model):
    """load deepsparse"""

    from huggingface_hub import snapshot_download  # pylint: disable=import-outside-toplevel

    # Download the model from HF
    model_folder = snapshot_download(repo_id=clip_model)
    # Compile the model with DeepSparse
    model = DeepSparseWrapper(model_path=model_folder)

    from deepsparse.clip.constants import CLIP_RGB_MEANS, CLIP_RGB_STDS  # pylint: disable=import-outside-toplevel

    def process_image(image):
        image = model.visual_model._preprocess_transforms(image.convert("RGB"))  # pylint: disable=protected-access
        image_array = np.array(image)
        image_array = image_array.transpose(2, 0, 1).astype("float32")
        image_array /= 255.0
        image_array = (image_array - np.array(CLIP_RGB_MEANS).reshape((3, 1, 1))) / np.array(CLIP_RGB_STDS).reshape(
            (3, 1, 1)
        )
        return torch.from_numpy(np.ascontiguousarray(image_array, dtype=np.float32))

    return model, process_image


@lru_cache(maxsize=None)
def get_tokenizer(clip_model):
    """Load clip"""
    if clip_model.startswith("open_clip:"):
        import open_clip  # pylint: disable=import-outside-toplevel

        clip_model = clip_model[len("open_clip:") :]
        return open_clip.get_tokenizer(clip_model)
    else:
        return lambda t: clip.tokenize(t, truncate=True)


@lru_cache(maxsize=None)
def load_clip_without_warmup(clip_model, use_jit, device, clip_cache_path):
    """Load clip"""
    if clip_model.startswith("open_clip:"):
        clip_model = clip_model[len("open_clip:") :]
        model, preprocess = load_open_clip(clip_model, use_jit, device, clip_cache_path)
    elif clip_model.startswith("hf_clip:"):
        clip_model = clip_model[len("hf_clip:") :]
        model, preprocess = load_hf_clip(clip_model, device)
    elif clip_model.startswith("nm:"):
        clip_model = clip_model[len("nm:") :]
        model, preprocess = load_deepsparse(clip_model)
    else:
        model, preprocess = clip.load(clip_model, device=device, jit=use_jit, download_root=clip_cache_path)
    return model, preprocess


@lru_cache(maxsize=None)
def load_clip(clip_model="ViT-B/32", use_jit=True, warmup_batch_size=1, clip_cache_path=None, device=None):
    """Load clip then warmup"""
    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    model, preprocess = load_clip_without_warmup(clip_model, use_jit, device, clip_cache_path)

    start = time.time()
    print(f"warming up with batch size {warmup_batch_size} on {device}", flush=True)
    warmup(warmup_batch_size, device, preprocess, model)
    duration = time.time() - start
    print(f"done warming up in {duration}s", flush=True)
    return model, preprocess


def warmup(batch_size, device, preprocess, model):
    fake_img = Image.new("RGB", (224, 224), color="red")
    fake_text = ["fake"] * batch_size
    image_tensor = torch.cat([torch.unsqueeze(preprocess(fake_img), 0)] * batch_size).to(device)
    text_tokens = clip.tokenize(fake_text).to(device)
    for _ in range(2):
        with torch.no_grad():
            model.encode_image(image_tensor)
            model.encode_text(text_tokens)



"""
Read environment variables and distribute work work to this rank.
This script is called by the slurm distributor.

Will launch a worker.
"""

import os
import json

from torch.cuda import set_device

from clip_retrieval.clip_inference.worker import worker


def get_task_list(num_tasks, world_size, global_rank, local_rank):
    """Get the list of tasks to process."""
    tasks_per_worker = num_tasks // world_size

    # Assign a subset of tasks to each worker
    start = global_rank * tasks_per_worker
    end = start + tasks_per_worker

    # If the tasks don't divide evenly
    # then we should redistribute the remainder
    if global_rank < num_tasks % world_size:
        start += global_rank
        end += global_rank + 1
    else:
        start += num_tasks % world_size
        end += num_tasks % world_size

    tasks = list(range(start, end))

    print(f"worker global rank:{global_rank}\tlocal rank: {local_rank}\tprocessing tasks {tasks}")

    return tasks


def slurm_worker():
    """Distribute work to this job and launch a job."""

    # Read environment variables
    # These are set by slurm_distributor or SLURM itself
    num_tasks = int(os.environ["NUM_TASKS"])
    global_rank = int(os.environ["SLURM_PROCID"])
    world_size = int(os.environ["WORLD_SIZE"])
    local_rank = int(os.environ["SLURM_LOCALID"])

    # Read the worker args from the file
    with open(os.environ["WORKER_ARGS_PATH"], "r", encoding="utf-8") as worker_args_file:
        worker_args = json.load(worker_args_file)

    # Find the range of tasks to process
    tasks = get_task_list(num_tasks, world_size, global_rank, local_rank)

    # set device
    set_device(local_rank)

    # Launch the worker
    worker(tasks, **worker_args)


if __name__ == "__main__":
    slurm_worker()



"""Clip client is a simple python module that allows you to query the backend remotely."""

import base64
import enum
import json
from pathlib import Path
from typing import Dict, List, Optional

import requests


class Modality(enum.Enum):
    IMAGE = "image"
    TEXT = "text"


class ClipClient:
    """Remotely query the CLIP backend via REST"""

    def __init__(
        self,
        url: str,
        indice_name: str,
        use_mclip: bool = False,
        aesthetic_score: int = 9,
        aesthetic_weight: float = 0.5,
        modality: Modality = Modality.IMAGE,
        num_images: int = 40,
        deduplicate: bool = True,
        use_safety_model: bool = True,
        use_violence_detector: bool = True,
    ):
        """
        url: (required) URL of the backend.
        indice_name: (required) which indice to search over e.g. "laion5B" or "laion_400m".
        use_mclip: (optional) whether to use mclip, a multilingual version of clip. Default is False.
        aesthetic_score: (optional) ranking score for aesthetic, higher is prettier. Default is 9.
        aesthetic_weight: (optional) weight of the aesthetic score, between 0 and 1. Default is 0.5.
        modality: (optional) Search modality. One of Modality.IMAGE or Modality.TEXT. Default is Modality.IMAGE.
        num_images: (optional) Number of images to return. Default is 40.
        deduplicate: (optional) Whether to deduplicate the result by image embedding. Default is true.
        use_safety_model: (optional) Whether to remove unsafe images. Default is true.
        use_violence_detector: (optional) Whether to remove images with violence. Default is true.
        """
        self.url = url
        self.indice_name = indice_name
        self.use_mclip = use_mclip
        self.aesthetic_score = aesthetic_score
        self.aesthetic_weight = aesthetic_weight
        self.modality = modality.value
        self.num_images = num_images
        self.deduplicate = deduplicate
        self.use_safety_model = use_safety_model
        self.use_violence_detector = use_violence_detector

    def query(
        self,
        text: Optional[str] = None,
        image: Optional[str] = None,
        embedding_input: Optional[list] = None,
    ) -> List[Dict]:
        """
        Given text or image/s, search for other captions/images that are semantically similar.

        Args:
            text: text to be searched semantically.
            image: base64 string of image to be searched semantically

        Returns:
            List of dictionaries containing the results in the form of:
            [
                {
                    "id": 42,
                    "similarity": 0.323424523424,
                    "url": "https://example.com/image.jpg",
                    "caption": "This is a caption",
                },
                ...
            ]
        """
        if text and image:
            raise ValueError("Only one of text or image can be provided.")
        if text:
            return self.__search_knn_api__(text=text)
        elif image:
            if image.startswith("http"):
                return self.__search_knn_api__(image_url=image)
            else:
                assert Path(image).exists(), f"{image} does not exist."
                return self.__search_knn_api__(image=image)
        elif embedding_input:
            return self.__search_knn_api__(embedding_input=embedding_input)
        else:
            raise ValueError("Either text or image must be provided.")

    def __search_knn_api__(
        self,
        text: Optional[str] = None,
        image: Optional[str] = None,
        image_url: Optional[str] = None,
        embedding_input: Optional[list] = None,
    ) -> List:
        """
        This function is used to send the request to the knn service.
        It represents a direct API call and should not be called directly outside the package.

        Args:
            text: text to be searched semantically.
            image: base64 string of image to be searched semantically.
            image_url: url of the image to be searched semantically.
            embedding_input: embedding input

        Returns:
            List of dictionaries containing the results in the form of:
            [
                {
                    "id": 42,
                    "similarity": 0.323424523424,
                    "url": "https://example.com/image.jpg",
                    "caption": "This is a caption",
                },
                ...
            ]

        """
        if image:
            # Convert image to base64 string
            with open(image, "rb") as image_file:
                encoded_string = base64.b64encode(image_file.read())
                image = str(encoded_string.decode("utf-8"))
        return requests.post(
            self.url,
            data=json.dumps(
                {
                    "text": text,
                    "image": image,
                    "image_url": image_url,
                    "embedding_input": embedding_input,
                    "deduplicate": self.deduplicate,
                    "use_safety_model": self.use_safety_model,
                    "use_violence_detector": self.use_violence_detector,
                    "indice_name": self.indice_name,
                    "use_mclip": self.use_mclip,
                    "aesthetic_score": self.aesthetic_score,
                    "aesthetic_weight": self.aesthetic_weight,
                    "modality": self.modality,
                    "num_images": self.num_images,
                    # num_results_ids is hardcoded to the num_images parameter.
                    "num_result_ids": self.num_images,
                }
            ),
        ).json()



"""Reader module provides files and webdataset readers"""

from pathlib import Path
from PIL import Image
from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
import io


def folder_to_keys(folder, enable_text=True, enable_image=True, enable_metadata=False):
    """returns a list of keys from a folder of images and text"""
    path = Path(folder)
    text_files = None
    metadata_files = None
    image_files = None
    if enable_text:
        text_files = [*path.glob("**/*.txt")]
        text_files = {text_file.stem: text_file for text_file in text_files}
    if enable_image:
        image_files = [
            *path.glob("**/*.png"),
            *path.glob("**/*.jpg"),
            *path.glob("**/*.jpeg"),
            *path.glob("**/*.bmp"),
            *path.glob("**/*.webp"),
        ]
        image_files = {image_file.stem: image_file for image_file in image_files}
    if enable_metadata:
        metadata_files = [*path.glob("**/*.json")]
        metadata_files = {metadata_file.stem: metadata_file for metadata_file in metadata_files}

    keys = None
    join = lambda new_set: new_set & keys if keys is not None else new_set
    if enable_text:
        keys = join(text_files.keys())
    elif enable_image:
        keys = join(image_files.keys())
    elif enable_metadata:
        keys = join(metadata_files.keys())

    keys = list(sorted(keys))

    return keys, text_files, image_files, metadata_files


def get_image_dataset():
    """retrieve image dataset module without importing torch at the top level"""

    from torch.utils.data import Dataset  # pylint: disable=import-outside-toplevel

    class ImageDataset(Dataset):
        """ImageDataset is a pytorch Dataset exposing image and text tensors from a folder of image and text"""

        def __init__(
            self,
            preprocess,
            folder,
            enable_text=True,
            enable_image=True,
            enable_metadata=False,
            input_sampler=lambda a: a,
        ):
            super().__init__()
            import clip  # pylint: disable=import-outside-toplevel

            self.keys, text_files, image_files, metadata_files = folder_to_keys(
                folder, enable_text, enable_image, enable_metadata
            )
            self.keys = input_sampler(self.keys)
            self.enable_text = enable_text
            self.enable_image = enable_image
            self.enable_metadata = enable_metadata
            keys_set = set(self.keys)
            if self.enable_text:
                self.tokenizer = lambda text: clip.tokenize([text], truncate=True)[0]
                self.text_files = {k: v for k, v in text_files.items() if k in keys_set}
            if self.enable_image:
                self.image_files = {k: v for k, v in image_files.items() if k in keys_set}
                self.image_transform = preprocess
            if self.enable_metadata:
                self.metadata_files = {k: v for k, v in metadata_files.items() if k in keys_set}

        def __len__(self):
            return len(self.keys)

        def __getitem__(self, ind):
            key = self.keys[ind]
            output = {}

            if self.enable_image:
                image_file = self.image_files[key]
                image_tensor = self.image_transform(Image.open(image_file))
                output["image_filename"] = str(image_file)
                output["image_tensor"] = image_tensor

            if self.enable_text:
                text_file = self.text_files[key]
                caption = text_file.read_text()
                tokenized_text = self.tokenizer(caption)
                output["text_tokens"] = tokenized_text
                output["text"] = caption

            if self.enable_metadata:
                metadata_file = self.metadata_files[key]
                metadata = metadata_file.read_text()
                output["metadata"] = metadata

            return output

    return ImageDataset


def create_webdataset(
    urls,
    image_transform,
    enable_text=True,
    enable_image=True,
    image_key="jpg",
    caption_key="txt",
    enable_metadata=False,
    cache_path=None,
    input_sampler=lambda a: a,
):
    """Create a WebDataset reader, it can read a webdataset of image, text and json"""
    import clip  # pylint: disable=import-outside-toplevel
    import webdataset as wds  # pylint: disable=import-outside-toplevel

    urls = input_sampler(urls)

    dataset = wds.WebDataset(urls, cache_dir=cache_path, cache_size=10**10, handler=wds.handlers.warn_and_continue)
    tokenizer = lambda text: clip.tokenize([text], truncate=True)[0]

    def filter_dataset(item):
        if enable_text and caption_key not in item:
            return False
        if enable_image and image_key not in item:
            return False
        if enable_metadata and "json" not in item:
            return False
        return True

    filtered_dataset = dataset.select(filter_dataset)

    def preprocess_dataset(item):
        output = {}
        if enable_image:
            image_data = item[image_key]
            image = Image.open(io.BytesIO(image_data))
            image_tensor = image_transform(image)
            output["image_filename"] = item["__key__"]
            output["image_tensor"] = image_tensor

        if enable_text:
            text = item[caption_key]
            caption = text.decode("utf-8")
            tokenized_text = tokenizer(caption)
            output["text_tokens"] = tokenized_text
            output["text"] = caption

        if enable_metadata:
            metadata_file = item["json"]
            metadata = metadata_file.decode("utf-8")
            output["metadata"] = metadata
        return output

    transformed_dataset = filtered_dataset.map(preprocess_dataset, handler=wds.handlers.warn_and_continue)
    return transformed_dataset


def dataset_to_dataloader(dataset, batch_size, num_prepro_workers, input_format):
    """Create a pytorch dataloader from a dataset"""

    def collate_fn(batch):
        batch = list(filter(lambda x: x is not None, batch))
        return default_collate(batch)

    data = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_prepro_workers,
        pin_memory=True,
        prefetch_factor=2,
        collate_fn=collate_fn if input_format == "files" else None,
    )
    return data


class FilesReader:
    """FilesReader is a reader that reads files from a folder"""

    def __init__(
        self,
        sampler,
        preprocess,
        input_dataset,
        batch_size,
        num_prepro_workers,
        enable_text=True,
        enable_image=True,
        enable_metadata=False,
    ) -> None:
        super().__init__()
        dataset = get_image_dataset()(preprocess, input_dataset, enable_text, enable_image, enable_metadata, sampler)
        self.dataloader = dataset_to_dataloader(dataset, batch_size, num_prepro_workers, "files")

    def __iter__(self):
        for batch in self.dataloader:
            yield batch


class WebdatasetReader:
    """WebdatasetReader is a reader that reads samples from a webdataset"""

    def __init__(
        self,
        sampler,
        preprocess,
        input_dataset,
        batch_size,
        num_prepro_workers,
        enable_text=True,
        enable_image=True,
        enable_metadata=False,
        wds_image_key="jpg",
        wds_caption_key="txt",
        cache_path=None,
    ):
        self.batch_size = batch_size
        dataset = create_webdataset(
            input_dataset,
            preprocess,
            enable_text=enable_text,
            enable_image=enable_image,
            image_key=wds_image_key,
            caption_key=wds_caption_key,
            enable_metadata=enable_metadata,
            cache_path=cache_path,
            input_sampler=sampler,
        )
        self.dataloader = dataset_to_dataloader(dataset, batch_size, num_prepro_workers, "webdataset")

    def __iter__(self):
        for batch in self.dataloader:
            yield batch



"""main module combines distributor, runner, reader, mapper, writer to produce clip embeddings"""

import fire
import math
from braceexpand import braceexpand

from clip_retrieval.clip_inference.logger import LoggerReader
from clip_retrieval.clip_inference.reader import folder_to_keys
from clip_retrieval.clip_inference.slurm_distributor import SlurmDistributor
from clip_retrieval.clip_inference.distributor import PysparkDistributor, SequentialDistributor


def calculate_partition_count(
    input_format,
    input_dataset,
    enable_image,
    enable_text,
    enable_metadata,
    write_batch_size,
    wds_number_file_per_input_file,
):
    """
    Calculate the partition count needed to store the resulting embeddings.

    Return:
        - the output partition count and the updated toggles for image, text and metadata.
    """

    sample_count = 0

    if input_format == "files":
        keys, text_files, image_files, metadata_files = folder_to_keys(
            input_dataset,
            enable_text=enable_text,
            enable_image=enable_image,
            enable_metadata=enable_metadata,
        )
        if text_files is None or len(text_files) == 0:
            enable_text = False
        if image_files is None or len(image_files) == 0:
            enable_image = False
        if metadata_files is None or len(metadata_files) == 0:
            enable_metadata = False
        if not enable_text and not enable_image and not enable_metadata:
            raise ValueError("no sample found")
        keys, text_files, image_files, metadata_files = folder_to_keys(
            input_dataset,
            enable_text=enable_text,
            enable_image=enable_image,
            enable_metadata=enable_metadata,
        )
        sample_count = len(keys)
    elif input_format == "webdataset":
        sample_count = len(input_dataset) * wds_number_file_per_input_file
    else:
        raise ValueError(f"Unsupported input_format {input_format}")

    if sample_count == 0:
        raise ValueError("no sample found")

    print(f"The number of samples has been estimated to be {sample_count}")

    output_partition_count = math.ceil(sample_count / write_batch_size)

    return output_partition_count, enable_text, enable_image, enable_metadata


# pylint: disable=unused-argument
def main(
    input_dataset,
    output_folder,
    input_format="files",
    cache_path=None,
    batch_size=256,
    num_prepro_workers=4,
    enable_text=True,
    enable_image=True,
    enable_metadata=False,
    write_batch_size=10**6,
    wds_image_key="jpg",
    wds_caption_key="txt",
    clip_model="ViT-B/32",
    mclip_model="sentence-transformers/clip-ViT-B-32-multilingual-v1",
    use_mclip=False,
    use_jit=False,
    distribution_strategy="sequential",
    wds_number_file_per_input_file=10000,
    output_partition_count=None,
    wandb_project="clip_retrieval",
    enable_wandb=False,
    clip_cache_path=None,
    slurm_job_name=None,
    slurm_partition=None,
    slurm_nodes=None,
    slurm_job_comment=None,
    slurm_nodelist=None,
    slurm_exclude=None,
    slurm_job_timeout=None,
    slurm_cache_path=None,
    slurm_verbose_wait=False,
):
    # package arguments to pass on to the distributor
    local_args = dict(locals())

    expanded_dataset = list(braceexpand(input_dataset)) if input_format == "webdataset" else input_dataset

    # compute this now for the distributors to use
    if output_partition_count is None:
        output_partition_count, enable_text, enable_image, enable_metadata = calculate_partition_count(
            input_format=input_format,
            input_dataset=expanded_dataset,
            enable_image=enable_image,
            enable_text=enable_text,
            enable_metadata=enable_metadata,
            write_batch_size=write_batch_size,
            wds_number_file_per_input_file=wds_number_file_per_input_file,
        )

        # update the local args to match the computed values
        local_args["output_partition_count"] = output_partition_count
        local_args["enable_text"] = enable_text
        local_args["enable_image"] = enable_image
        local_args["enable_metadata"] = enable_metadata

    local_args.pop("wds_number_file_per_input_file")
    local_args.pop("write_batch_size")
    local_args.pop("distribution_strategy")
    local_args.pop("wandb_project")
    local_args.pop("enable_wandb")

    tasks = list(range(output_partition_count))
    worker_args = {k: v for k, v in local_args.items() if not k.startswith("slurm_")}

    if distribution_strategy == "sequential":
        distributor = SequentialDistributor(tasks=tasks, worker_args=worker_args)
    elif distribution_strategy == "pyspark":
        distributor = PysparkDistributor(tasks=tasks, worker_args=worker_args)
    elif distribution_strategy == "slurm":
        slurm_args = {k.lstrip("slurm_"): v for k, v in local_args.items() if k.startswith("slurm_")}
        distributor = SlurmDistributor(tasks=tasks, worker_args=worker_args, slurm_args=slurm_args)
    else:
        print(
            f"The {distribution_strategy} strategy is not implemented. Please choose from: [sequential, pyspark, slurm]"
        )

    logger_reader = LoggerReader(
        stats_folder=output_folder + "/stats",
        wandb_project=wandb_project,
        enable_wandb=enable_wandb,
    )

    logger_reader.start()

    distributor()

    logger_reader.end()


if __name__ == "__main__":
    fire.Fire(main)



"""mapper module transform images and text to embeddings"""

import torch
from clip_retrieval.load_clip import load_clip
from sentence_transformers import SentenceTransformer


def normalized(a, axis=-1, order=2):
    import numpy as np  # pylint: disable=import-outside-toplevel

    l2 = np.atleast_1d(np.linalg.norm(a, order, axis))
    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis)


class ClipMapper:
    """transforms images and texts into clip embeddings"""

    def __init__(
        self,
        enable_image,
        enable_text,
        enable_metadata,
        use_mclip,
        clip_model,
        use_jit,
        mclip_model,
        warmup_batch_size=1,
        clip_cache_path=None,
    ):
        self.enable_image = enable_image
        self.enable_text = enable_text
        self.enable_metadata = enable_metadata
        self.use_mclip = use_mclip
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        model, _ = load_clip(
            clip_model=clip_model, use_jit=use_jit, warmup_batch_size=warmup_batch_size, clip_cache_path=clip_cache_path
        )
        self.model_img = model.encode_image
        self.model_txt = model.encode_text
        if use_mclip:
            print("\nLoading MCLIP model for text embedding\n")
            mclip = SentenceTransformer(mclip_model)
            self.model_txt = mclip.encode

    def __call__(self, item):
        with torch.no_grad():
            image_embs = None
            text_embs = None
            image_filename = None
            text = None
            metadata = None
            if self.enable_image:
                image_features = self.model_img(item["image_tensor"].to(self.device))
                image_features /= image_features.norm(dim=-1, keepdim=True)
                image_embs = image_features.cpu().to(torch.float16).numpy()
                image_filename = item["image_filename"]
            if self.enable_text:
                if self.use_mclip:
                    text_embs = normalized(self.model_txt(item["text"]))
                else:
                    text_features = self.model_txt(item["text_tokens"].to(self.device))
                    text_features /= text_features.norm(dim=-1, keepdim=True)
                    text_embs = text_features.cpu().to(torch.float16).numpy()
                text = item["text"]
            if self.enable_metadata:
                metadata = item["metadata"]

            return {
                "image_embs": image_embs,
                "text_embs": text_embs,
                "image_filename": image_filename,
                "text": text,
                "metadata": metadata,
            }



"""
Inference Worker:

A completely independent process that will be started once for each GPU node.
Distributors will call this either through the CLI or directly.

The worker sequentially process the tasks passed to it.
Tasks are lists of partition_id's that this worker will be responsible for.
"""

import fire
from braceexpand import braceexpand

from clip_retrieval.clip_inference.runner import Runner
from clip_retrieval.clip_inference.mapper import ClipMapper
from clip_retrieval.clip_inference.writer import NumpyWriter
from clip_retrieval.clip_inference.logger import LoggerWriter
from clip_retrieval.clip_inference.reader import FilesReader, WebdatasetReader
from clip_retrieval.load_clip import load_clip


def worker(
    tasks,
    input_dataset,
    output_folder,
    output_partition_count,
    input_format="files",
    cache_path=None,
    batch_size=256,
    num_prepro_workers=4,
    enable_text=True,
    enable_image=True,
    enable_metadata=False,
    wds_image_key="jpg",
    wds_caption_key="txt",
    clip_model="ViT-B/32",
    mclip_model="sentence-transformers/clip-ViT-B-32-multilingual-v1",
    use_mclip=False,
    use_jit=True,
    clip_cache_path=None,
):
    """Start a worker"""
    print("Starting the worker", flush=True)

    # check for brace expansion
    if input_format == "webdataset" and not isinstance(input_dataset, list):
        input_dataset = list(braceexpand(input_dataset))

    print(f"dataset is {len(input_dataset)}", flush=True)

    def reader_builder(sampler):
        _, preprocess = load_clip(
            clip_model=clip_model, use_jit=use_jit, warmup_batch_size=batch_size, clip_cache_path=clip_cache_path
        )
        if input_format == "files":
            return FilesReader(
                sampler,
                preprocess,
                input_dataset,
                batch_size,
                num_prepro_workers,
                enable_text=enable_text,
                enable_image=enable_image,
                enable_metadata=enable_metadata,
            )
        elif input_format == "webdataset":
            return WebdatasetReader(
                sampler,
                preprocess,
                input_dataset,
                batch_size,
                num_prepro_workers,
                enable_text=enable_text,
                enable_image=enable_image,
                enable_metadata=enable_metadata,
                wds_image_key=wds_image_key,
                wds_caption_key=wds_caption_key,
                cache_path=cache_path,
            )
        else:
            raise ValueError(f"Unknown input_format: {input_format}")

    def mapper_builder():
        return ClipMapper(
            enable_image=enable_image,
            enable_text=enable_text,
            enable_metadata=enable_metadata,
            use_mclip=use_mclip,
            clip_model=clip_model,
            use_jit=use_jit,
            mclip_model=mclip_model,
            clip_cache_path=clip_cache_path,
            warmup_batch_size=batch_size,
        )

    def writer_builder(i):
        return NumpyWriter(
            partition_id=i,
            output_folder=output_folder,
            enable_text=enable_text,
            enable_image=enable_image,
            enable_metadata=enable_metadata,
            output_partition_count=output_partition_count,
        )

    def logger_builder(i):
        return LoggerWriter(
            partition_id=i,
            stats_folder=output_folder + "/stats",
        )

    runner = Runner(
        reader_builder=reader_builder,
        mapper_builder=mapper_builder,
        writer_builder=writer_builder,
        logger_builder=logger_builder,
        output_partition_count=output_partition_count,
    )

    for task in tasks:
        print(f"Starting work on task {task}", flush=True)
        runner(task)


if __name__ == "__main__":
    fire.Fire(worker)



"""Distribute work using SLURM"""

import os
import time
import json
import subprocess
from datetime import datetime

TIMESTAMP = datetime.now().timestamp()


class SlurmDistributor:
    """distribute work across a collection of slurm jobs"""

    def __init__(self, tasks, worker_args, slurm_args):
        self.num_tasks = len(tasks)
        self.worker_args = worker_args
        self.slurm_args = slurm_args

        self.job_timeout = slurm_args.pop("job_timeout")
        self.verbose_wait = slurm_args.pop("verbose_wait")

    def __call__(self):
        """
        Create a sbatch file, submit it to slurm, and wait for it to finish.
        """
        # pop the cache path from the slurm args to remove it
        cache_path = self.slurm_args.pop("cache_path")

        # create the cache path if it doesn't exist
        if cache_path is None:
            cache_path = os.path.expanduser("~/.cache")

        os.makedirs(cache_path, exist_ok=True)

        # make the filenames unique using the current timestamp
        sbatch_script_path = os.path.join(cache_path, f"sbatch_script_{TIMESTAMP}.sh")

        # save the file to the cache path
        with open(sbatch_script_path, "w", encoding="utf-8") as sbatch_file:
            sbatch_file.write(
                self._generate_sbatch(cache_path=cache_path, slurm_args=self.slurm_args, worker_args=self.worker_args)
            )

        # now we need to run the job
        status = self._run_job(sbatch_script_path)

        # interpret the results
        if status == "success":
            print("job succeeded")
            return True
        elif status == "failed":
            print("job failed")
            return False
        else:
            print("exception occurred")
            return False

    def _run_job(self, sbatch_file):
        """
        Run a job and wait for it to finish.
        """
        try:
            job_id = self._start_job(sbatch_file)

            print(f"waiting for job {job_id}")

            timeout = self.job_timeout

            if timeout is None:
                print("You have not specified a timeout, defaulting to 2 weeks.")
                timeout = 1.21e6

            status = self._wait_for_job_to_finish(job_id=job_id, timeout=timeout)

            if not status:
                print(f"canceling {job_id}")
                subprocess.check_output(["scancel", job_id]).decode("utf8")
                status = self._wait_for_job_to_finish(job_id)
                print("job cancelled")
                return "failed"
            else:
                print("job succeeded")
                return "success"
        except Exception as e:  # pylint: disable=broad-except
            print(e)
            return "exception occurred"

    def _wait_for_job_to_finish(self, job_id, timeout=30):
        t = time.time()
        while 1:
            if time.time() - t > timeout:
                return False
            time.sleep(1)
            if self._is_job_finished(job_id):
                return True

    def _is_job_finished(self, job_id):
        status = subprocess.check_output(["squeue", "-j", job_id]).decode("utf8")

        if self.verbose_wait:
            print(f"job status is {status}")

        return status == "slurm_load_jobs error: Invalid job id specified" or len(status.split("\n")) == 2

    def _start_job(self, sbatch_file):
        """start job"""
        args = ["sbatch"]
        args.append(sbatch_file)
        sbatch_output = subprocess.check_output(args).decode("utf8")
        lines = sbatch_output.split("\n")

        lines = [line for line in lines if "Submitted" in line]
        if len(lines) == 0:
            raise ValueError(f"slurm sbatch failed: {sbatch_output}")

        parsed_sbatch = lines[0].split(" ")
        job_id = parsed_sbatch[3].strip()
        return job_id

    def _write_json_worker_args(self, worker_args, cache_path):
        """write the worker args to a json file"""
        worker_args_path = os.path.join(cache_path, f"worker_args_{TIMESTAMP}.json")
        with open(worker_args_path, "w", encoding="utf-8") as worker_args_file:
            json.dump(worker_args, worker_args_file, indent=4)
        return worker_args_path

    def _generate_sbatch(self, cache_path, slurm_args, worker_args):
        """
        Generate sbatch for a worker.

        sbatch: allows you to specify a configuration and task in a file
            - https://slurm.schedmd.com/sbatch.html
        """
        # write the worker args to a file
        worker_args_path = self._write_json_worker_args(worker_args, cache_path)

        venv = os.environ["VIRTUAL_ENV"]
        scomment = ("--comment " + slurm_args["job_comment"]) if ["job_comment"] is not None else ""
        sbatch_scomment = (
            ("#SBATCH --comment " + slurm_args["job_comment"]) if slurm_args["job_comment"] is not None else ""
        )
        nodelist = ("#SBATCH --nodelist " + slurm_args["nodelist"]) if slurm_args["nodelist"] is not None else ""
        exclude = ("#SBATCH --exclude " + slurm_args["exclude"]) if slurm_args["exclude"] is not None else ""

        return f"""#!/bin/bash
# Define sbatch config, use exclusive to capture all resources in each node
#SBATCH --partition={slurm_args["partition"]}
#SBATCH --job-name={slurm_args["job_name"]}
#SBATCH --output={cache_path}/slurm-%x_%j.out
#SBATCH --nodes={slurm_args["nodes"]}
#SBATCH --ntasks-per-node=8
#SBATCH --cpus-per-gpu=6
#SBATCH --gres=gpu:8
#SBATCH --exclusive

{sbatch_scomment}
{nodelist}
{exclude}

# Environment variables for the inner script
export NUM_TASKS={self.num_tasks}
export WORLD_SIZE={slurm_args["nodes"] * 8} # 8 gpus per node
export WORKER_ARGS_PATH={worker_args_path}

# Run the internal script
source {venv}/bin/activate
srun --cpu_bind=v --accel-bind=gn {scomment} clip-retrieval inference.slurm_worker
"""



"""The logger module allows logging to stdout and wandb"""

from collections import defaultdict
import fsspec
import multiprocessing
import time
import json
import wandb
import queue
import traceback


class LoggerWriter:
    """the logger writer write stats to json file, for each worker"""

    def __init__(self, partition_id, stats_folder):
        self.partition_id = partition_id
        self.stats_folder = stats_folder

    def start(self):
        ctx = multiprocessing.get_context("spawn")
        self.queue = ctx.Queue()
        self.updater_process = ctx.Process(target=self.updater)
        self.updater_process.start()

    def end(self):
        self.queue.put(None)
        self.updater_process.join()
        self.queue.close()

    def __call__(self, stats):
        self.queue.put(stats)

    def updater(self):
        """updater process that writes stats to file from the queue"""
        stats = defaultdict(lambda: 0)
        fs, relative_path = fsspec.core.url_to_fs(self.stats_folder)
        last_write = None
        while True:
            item = self.queue.get()
            if item is None:
                self.write_stats(stats, fs, relative_path, False)
                return
            for k in item:
                stats[k] += item[k]
            if last_write is None or time.time() - last_write > 5:
                self.write_stats(stats, fs, relative_path, True)
                last_write = time.time()

    def sum(self, stats, new_stats):
        for k in stats.keys():
            stats[k] += new_stats[k]
        return stats

    def write_stats(self, stats, fs, relative_path, wip):
        fs.makedirs(relative_path, exist_ok=True)
        if not wip and fs.exists(relative_path + f"/wip_{self.partition_id}.json"):
            fs.rm(relative_path + f"/wip_{self.partition_id}.json")
        prefix = "wip_" if wip else ""
        with fs.open(relative_path + f"/{prefix}{self.partition_id}.json", "w") as f:
            f.write(json.dumps(stats))


class LoggerReader:
    """the logger reader read stats of all json files and aggregate them"""

    def __init__(self, stats_folder, wandb_project="clip_retrieval", enable_wandb=False):
        self.stats_folder = stats_folder
        self.enable_wandb = enable_wandb
        self.wandb_project = wandb_project
        self.log_interval = 5

    def start(self):
        ctx = multiprocessing.get_context("spawn")
        self.queue = ctx.Queue()
        self.start_time = time.perf_counter()
        self.reader_process = ctx.Process(target=self.reader)
        self.reader_process.start()

    def end(self):
        self.queue.put("end")
        self.reader_process.join()
        self.queue.close()

    def reader(self):
        """reader process that reads stats from files and aggregates them"""
        try:  # pylint: disable=too-many-nested-blocks
            if self.enable_wandb:
                self.current_run = wandb.init(project=self.wandb_project)
            else:
                self.current_run = None

            last_check = 0
            stats = {}
            start_time_no_initial_load = float("inf")
            fs, relative_path = fsspec.core.url_to_fs(self.stats_folder, use_listings_cache=False)

            fs.makedirs(relative_path, exist_ok=True)

            while True:  # pylint: disable=too-many-nested-blocks
                time.sleep(0.1)
                try:
                    self.queue.get(False)
                    last_one = True
                except queue.Empty as _:
                    last_one = False
                if not last_one and time.perf_counter() - last_check < self.log_interval:
                    continue

                last_check = time.perf_counter()

                stats_files = fs.glob(relative_path + "/*.json")
                for k in stats_files:
                    filename = k.split("/")[-1]
                    if filename[:4] == "wip_" or filename not in stats:
                        for i in range(5):  # pylint: disable=unused-variable
                            try:
                                fs.invalidate_cache()
                                if not fs.exists(k):
                                    continue
                                with fs.open(k, "r") as f:
                                    stats[filename] = json.loads(f.read())
                                if filename[:4] != "wip_" and "wip_" + filename in stats:
                                    del stats["wip_" + filename]
                                break
                            except Exception as e:  # pylint: disable=broad-except
                                if i == 4:
                                    print(f"failed to read {k} error : {e}")
                                time.sleep(1)

                stats_aggregated = defaultdict(lambda: 0)
                for k, v in stats.items():
                    for k2 in v:
                        stats_aggregated[k2] += v[k2]

                for v in stats.values():
                    start_time_no_initial_load = min(start_time_no_initial_load, v["start_time"])

                current_time = time.perf_counter()
                current_real_time = time.time()
                total_duration = current_time - self.start_time

                if stats_aggregated["sample_count"] == 0:
                    if last_one:
                        self._finish()
                        break
                    continue
                total_duration_no_initial_load = current_real_time - start_time_no_initial_load

                stats_aggregated["average_read_duration_per_sample"] = (
                    stats_aggregated["read_duration"] / stats_aggregated["sample_count"]
                )
                stats_aggregated["average_inference_duration_per_sample"] = (
                    stats_aggregated["inference_duration"] / stats_aggregated["sample_count"]
                )
                stats_aggregated["average_write_duration_per_sample"] = (
                    stats_aggregated["write_duration"] / stats_aggregated["sample_count"]
                )
                stats_aggregated["average_total_duration_per_sample"] = (
                    stats_aggregated["total_duration"] / stats_aggregated["sample_count"]
                )
                stats_aggregated["sample_per_sec"] = stats_aggregated["sample_count"] / total_duration
                stats_aggregated["total_job_duration"] = total_duration
                stats_aggregated["total_duration_no_initial_load"] = total_duration_no_initial_load
                stats_aggregated["sample_per_sec_no_initial_load"] = (
                    stats_aggregated["sample_count"] / total_duration_no_initial_load
                )

                to_log = [
                    "sample_count",
                    "sample_per_sec",
                    "sample_per_sec_no_initial_load",
                    "total_job_duration",
                    "average_read_duration_per_sample",
                    "average_inference_duration_per_sample",
                    "average_write_duration_per_sample",
                    "average_total_duration_per_sample",
                ]
                stats_for_logging = {}
                for k in to_log:
                    stats_for_logging[k] = stats_aggregated[k]

                print(
                    "\r",
                    "sample_per_sec "
                    + str(int(stats_for_logging["sample_per_sec_no_initial_load"]))
                    + " ; sample_count "
                    + str(stats_for_logging["sample_count"])
                    + " ",
                    end="",
                )
                if self.enable_wandb:
                    wandb.log(stats_for_logging)

                if last_one:
                    self._finish()
                    break
        except Exception as e:  # pylint: disable=broad-except
            traceback.print_exc()
            print("logger error", e)
            self._finish()
            return

    def _finish(self):
        if self.current_run is not None:
            self.current_run.finish()



"""The runner combine reader, mapper and writer to produce clip embeddings"""

import time


class Sampler:
    """Sampler"""

    def __init__(self, output_partition_id, output_partition_count):
        self.output_partition_id = output_partition_id
        self.output_partition_count = output_partition_count

    def __call__(self, l):
        return [e for i, e in enumerate(l) if i % self.output_partition_count == self.output_partition_id]


class Runner:
    """Runner class"""

    def __init__(self, reader_builder, mapper_builder, writer_builder, logger_builder, output_partition_count):
        self.reader_builder = reader_builder
        self.mapper_builder = mapper_builder
        self.writer_builder = writer_builder
        self.logger_builder = logger_builder
        self.output_partition_count = output_partition_count

    def __call__(self, i):
        sampler = Sampler(i, self.output_partition_count)
        reader = self.reader_builder(sampler)
        writer = self.writer_builder(i)
        mapper = self.mapper_builder()
        logger = self.logger_builder(i)
        logger.start()
        iterator = reader.__iter__()
        while True:
            begin_time = time.time()
            start_time = time.perf_counter()
            try:
                batch = iterator.__next__()
            except StopIteration:
                break
            read_duration = time.perf_counter() - start_time
            start_time = time.perf_counter()
            embeddings = mapper(batch)
            inference_duration = time.perf_counter() - start_time
            start_time = time.perf_counter()
            writer(embeddings)
            write_duration = time.perf_counter() - start_time
            end_time = time.time()
            logger(
                {
                    "start_time": begin_time,
                    "end_time": end_time,
                    "read_duration": read_duration,
                    "inference_duration": inference_duration,
                    "write_duration": write_duration,
                    "total_duration": end_time - begin_time,
                    "sample_count": batch["image_tensor"].shape[0]
                    if "image_tensor" in batch
                    else batch["text_tokens"].shape[0],
                }
            )
        logger.end()
        writer.flush()
